
// JWS.java - The Java Web Server
// (C) 2007 Ben Cantrick

// package org.gully.mackys.JWS;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.concurrent.*;


/**
 * The Java Web Server.
 * A simple webserver written in Java. Possibly useful for learning purposes.
 * <br><br>For a quick demo, run the JWS with <tt>java JWS</tt> and then load
 * up <tt>http://localhost:8080</tt> to view the javadocs.
 *
 * @author Ben Cantrick
 * @version 2006/06/11
 */
public class JWS
{
    /**
     * Starts the JWS running by instantiating the JWSServer class.
     *
     * @param args  Command line arguments, ignored.
     */
    public static void main(String[] args)
    {
        JWSServer theServer = new JWSServer();
        theServer.RunServer();
    }
}


/**
 * Implementation backend class of the JWS.
 * JWSServer needs to be a different class than JWS because of issues
 * arising from the static-ness of main(). I even tried making JWSServer
 * an inner class within JWS, but javac still complained.
 */
class JWSServer
{

    // This is minimum and maximum number of threads in the worker
    // thread pool, and how long idle threads should be kept alive
    // when there are more than MINTHREADS in the pool.

    private final int WORKERPOOL_MINTHREADS = 5;
    private final int WORKERPOOL_MAXTHREADS = 50;
    private final int WORKERPOOL_KEEPALIVETIME_SECS = 60;
    private final int WORKERPOOL_TASKQ_MAXSIZE = 2 * WORKERPOOL_MAXTHREADS;

    // The port that the JWS listens for HTTP connections on.

    private final int JWS_LISTEN_PORT = 8080;

    /**
     * Inits and runs the server.
     * After initing all necessary data structures and worker threads,
     * this method continuously accepts connections on port JWS_LISTEN_PORT,
     * and places accepted connections into the work queue.
     */
    public void RunServer()
    {
        ThreadPoolExecutor workerPool = null;
        ServerSocket myServSock;
        Socket myClientSock;


        // Create and initialize the worker thread pool.
        try {
            workerPool = new ThreadPoolExecutor(
                WORKERPOOL_MINTHREADS,
                WORKERPOOL_MAXTHREADS,
                WORKERPOOL_KEEPALIVETIME_SECS,
                TimeUnit.SECONDS,
                new ArrayBlockingQueue<Runnable>(WORKERPOOL_TASKQ_MAXSIZE));
            workerPool.prestartAllCoreThreads();
        }
        catch (Exception e) {
            System.err.println("Can't init thread pool!!");
            e.printStackTrace();
            System.exit(-1);
        }

        // Start listening for connections.
        try {                             
            myServSock = new ServerSocket(JWS_LISTEN_PORT);
        }
        catch (IOException ioe)
        {
            myServSock = null;
            System.err.println("Can't listen()!! Exiting!");
            System.exit(-1);
        }

        while(true)
        {
            // Wait for and then accept a connection.
            try {                         
                myClientSock = myServSock.accept();
            }
            catch (IOException ioe)
            {
                System.err.println("Non-fatal error on accept(). Continuing.");
                myClientSock = null;
                continue;      // Go back to the top of the loop.
            }

            // Create a JWSWorker Runnable object on the socket, then hand
            // it over to the thread pool. The thread pool will run the
            // object's run() method when a worker thread becomes available
            // to run it.
            try {
                workerPool.execute(new JWSWorker(myClientSock));
            }

            catch (RejectedExecutionException rje) {
                System.err.println("execute() failed - worker thread problem? Closing socket w/error.");
                JWSWorker.sendErrorAndClose(myClientSock, null,
                    "HTTP/1.0 500 Internal Server Error\r\n\r\n" + 
                    JWSErrors.HTML500);
		myClientSock.close();
                myClientSock = null;
                continue;      // Go back to the top of the loop.
            }

        } // while(true)

    } // main()

} // class JWSServer



/**
 * A worker thread's data and run() method.
 * JWSWorker is not an actual thread in and of itself. The thread pool is
 * what maintains the threads for us. JWSWorker is just a Runnable class
 * that the thread pool will assign a thread to. That thread will execute
 * the JWSWorker object's run() method. Which will then process the HTTP
 * request coming in on the socket. When the JWSWorker's run() method
 * finishes, the thread pool will assign the thread to another JWSWorker
 * object. Which will process another HTTP request. And so on.
 */
class JWSWorker implements Runnable
{
    // The Java platform libs conveniently provide us with a
    // java.net.FileNameMap.getContentTypeFor(String filename) method.

    static public FileNameMap mimeData = URLConnection.getFileNameMap();


    // Per-instance data: the socket we're talking HTTP with.

    private Socket clientSock;


    /**
     * Don't use this constructor. When you create a JWSWorker object with
     * a null socket and put it into the task queue, at best you're
     * wasting memory and CPU cycles. At worst you may cause a null
     * pointer exception when a thread gets around to running the object.
     */
    private JWSWorker()
    {
    }

    /**
     * Create a JWSWorker object with a socket to operate on.
     *
     * @param  s  A connected socket to read the HTTP request from,
     *            and send the resulting data/error message back to.
     * @see       Socket
     */
    public JWSWorker(Socket s)
    {
        clientSock = s;
    }

    /**
     * run() is just a wrapper for handleConnection().
     * run() is executed by code inside the thread pool object. The
     * thread pool assigns a thread to run this object when such a
     * thread is available.
     */
    public void run()
    {
        handleConnection(clientSock);
    }

    /**
     * Reads in an HTTP request from the given {@link Socket} and sends
     * the appropriate reply. The request method and request URI are parsed
     * out of the HTTP request and the corresponding file contents are sent,
     * or an appropriate error code. Finally, the socket is closed.
     * <p>
     * <b>BUGS:</b> No security checking of the URI is performed, i.e. a
     * malicious user could request "/../../../etc/passwd" and get it.
     *
     * @param  clientSock  A connected socket to read the HTTP request from,
     *                     and send the resulting data/error message back to.
     * @see                Socket
     */
    private void handleConnection(Socket clientSock)
    {
        Scanner myScan;
        String httpMethod, requestURI;
        File myFileInfo;
        FileInputStream myFile;
        OutputStream sockOutStream;
        byte[] myByteBuf;
        int bytesRead;


        // Grab the HTTP request info and get ready to send... something.
        try {
            myScan = new Scanner( clientSock.getInputStream() );
        } catch (IOException ioe) {
            System.err.println("Non-fatal error on new Scanner().");
            return;
        }

        httpMethod = myScan.next();
        requestURI = myScan.next();

        // Process the requestURI, including index.html substitution.
       
        if(requestURI.charAt(0) == '/') {    // Strip leading /, if present.
            requestURI = requestURI.substring(1);
        }

        if(requestURI.length() == 0 || requestURI.endsWith("/")) {
            requestURI = requestURI + "index.html";
        }

        // All right, let's get ready to reply...

        try {
            sockOutStream = clientSock.getOutputStream();
        }
        catch(IOException ioe)
        {
            System.err.println("Non-fatal error on getOutputStream().");
            try { clientSock.close(); }
            catch(IOException innerIoe) { /* nothing */ }
            return;
        }

        // Might not have been a GET request...
        if(httpMethod.compareToIgnoreCase("GET") != 0) {
            sendErrorAndClose(clientSock, sockOutStream,
                "HTTP/1.0 501 Unimplemented\r\n\r\n" + JWSErrors.HTML501);
            return;
        }

        // Handle 403 and 404.
        myFileInfo = new File(requestURI);

        if(!myFileInfo.exists()) {
            sendErrorAndClose(clientSock, sockOutStream,
                "HTTP/1.0 404 Not Found\r\n\r\n" + JWSErrors.HTML404);
            return;
        }

        if(!myFileInfo.canRead()) {
            sendErrorAndClose(clientSock, sockOutStream,
                "HTTP/1.0 403 Permission Denied\r\n\r\n" + JWSErrors.HTML403);
            return;
        }

        // Open the file. Or not.
        try {
            myFile = new FileInputStream(requestURI);
        }
        catch(FileNotFoundException fnfe)
        {
            sendErrorAndClose(clientSock, sockOutStream,
                "HTTP/1.0 500 Internal Server Error\r\n\r\n" + JWSErrors.HTML500);
            return;
        }

        // Okay, send required headers and file contents.
        try {                              
            myByteBuf = new byte[4096];
            String headerLine;

            sockOutStream.write("HTTP/1.0 200 OK\r\n".getBytes());

            headerLine = "Content-Type: " + 
                mimeData.getContentTypeFor(requestURI) +
                "\r\n";
            sockOutStream.write(headerLine.getBytes());

            headerLine = "Content-Length: " + myFileInfo.length() + "\r\n\r\n";
            sockOutStream.write(headerLine.getBytes());

            while((bytesRead = myFile.read(myByteBuf)) != -1) {
                sockOutStream.write(myByteBuf, 0, bytesRead);
            }
        }
        catch(IOException ioe) {
            System.err.println("Non-fatal error on read() or write().");
        }
    
        // Shut 'r down and get out.
        try {
            sockOutStream.flush();            
            sockOutStream.close();
        }
        catch(IOException ioe) {
            System.err.println("Non-fatal error on close().");
        }

        try { clientSock.close(); }
        catch(IOException ioe) { /* nothing */ }

    } // handleConnection


    /**
     * Sends an error message and closes the connection. Catches IOException
     * on write(), flush() and close() but ignores them.
     *
     * @param sock     The socket to send the error to and close.
     * @param sockOut  The socket's output stream. Method creates if null.
     * @param error    A string that contains the error message.
     */
    public static void sendErrorAndClose(Socket sock,
        OutputStream sockOut, String error)
    {
        if(sockOut == null)
        {
            try {
                sockOut = sock.getOutputStream();
            }
            catch(IOException ioe)
            {
                System.err.println("Non-fatal error on getOutputStream().");
                try { sock.close(); }
                catch(IOException innerIoe) { /* nothing */ }
                return;
            }
        }

        try {
            sockOut.write(error.getBytes());
            sockOut.flush();
            sockOut.close();
        }
        catch(IOException ioe) { /* nothing */ }

        try{ sock.close(); }
        catch(IOException ioe) { /* nothing */ }

    }    // sendErrorAndClose()

}    // JWSWorker



/**
 * Some public static error messages for convenience.
 */
class JWSErrors {

    public static final String HTML403 = "<html><head><title>403 Permission Denied</title></head><body><h1>403 - Permission Denied</h1><p>You don't have permission to access that file, buddy!</p></body></html>\n";

    public static final String HTML404 = "<html><head><title>404 File Not Found</title></head><body><h1>404 - File Not Found</h1><p>Like, wow, man... I can't find that file, man. Like, totally, you know...</p></body></html>\n";

    public static final String HTML500 = "<html><head><title>500 Internal Server Error</title></head><body><h1>500 - The server encountered an internal error</h1><p>Oh, that's not good at all... is the server overloaded??</p></body></html>\n";

    public static final String HTML501 = "<html><head><title>501 Unimplemented</title></head><body><h1>501 - Unimplemented HTTP request method</h1><p>This server only understands HTTP GET requests.</p></body></html>\n";

}

